import argparse
import glob
import os
import time
import sys
from pycapo import CapoConfig
import logging
from pathlib import Path
from ._version import ___version___ as _version

from .pymygdala import LogHandler, SendNRAOEvent
from . import evlasciencedatamodel as sc
from . import NGASUtil as ng
from . import persistmetadata as pm
from . import manifestpersist as mp
from . import persistcaltables as pct
from . import persistimages as pimg
from . import archiveutils as au
from . import persistVLBAmetadata as vlbamd
from . import IngestionManifest as im
from .IngestionManifest import IngestionManifest

capoprofile = os.environ["CAPO_PROFILE"]

LOG = logging.getLogger("ingestion")

LOG.setLevel(logging.DEBUG)
handler = LogHandler(profile=capoprofile, application="Ingestion")
handler.setLevel(logging.DEBUG)
LOG.addHandler(handler)

ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
LOG.addHandler(ch)

_HELP_MESSAGE = """AAT Ingestion Utility, version {}:\n\tCreates metadata structures and places files into NGAS."""
ingest_parser = argparse.ArgumentParser(
    description=_HELP_MESSAGE.format(_version), formatter_class=argparse.RawTextHelpFormatter
)

__send_event = SendNRAOEvent(profile=capoprofile, application="SDM Ingestion")

_INGESTION_MANIFEST_FILENAME = "ingestion_manifest.json"
_SDM_INGESTION = "meta"
_BDF_INGESTION = "bdf"
_CALIBRATION_INGESTION = "cal"
_IMAGE_INGESTION = "img"
_VLBA_IDIFITS_INGESTION = "vlba"
_JSON_MANIFEST_INGESTION = "json"
_INVALID_INGESTION_OPERATION = "invalid"


def main():
    """
    Main method for ingesting files into metadata database and NGAS
    :return:
    :error returns
        -5: Invalid operation
    """

    # Primary flags indicating operation to perform
    mxg = ingest_parser.add_mutually_exclusive_group(required=True)
    mxg.add_argument("-b", "--bdf", help="Ingest BDFs for observation", action="store_true")
    mxg.add_argument("-c", "--cal", help="Ingest calibration products", action="store_true")
    mxg.add_argument("-i", "--img", help="Ingest imaging products and images", action="store_true")
    mxg.add_argument("-m", "--meta", help="Ingest science metadata for observation", action="store_true")
    mxg.add_argument("-v", "--vlba", help="Ingest VLBA science metadata", action="store_true")
    mxg.add_argument("-j", "--json", help="Ingestion using ingestion manifest", action="store_true")
    ingest_parser.add_argument("-p", "--path", help="Path to data")
    ingest_parser.add_argument(
        "-o", "--file", help="File (observation file set, calibration product tar file, BDF file, image)"
    )
    ingest_parser.add_argument("-g", "--filegroup", help="Calibration products parent eb or image project")
    ingest_parser.add_argument("-r", "--reingest", help="Re-ingest (delete) existing data", action="store_true")
    ingest_parser.add_argument("-n", "--nocal", help="Do not perform calibration", action="store_true")
    ingest_parser.add_argument(
        "-d", "--nongas", help="Only extract metadata, do not ingest into NGAS", action="store_true"
    )

    args = ingest_parser.parse_args()

    config = CapoConfig(profile=capoprofile)

    operation = getingestionoperation(args)
    if operation == _INVALID_INGESTION_OPERATION:
        LOG.critical(f"The operation is unspecified or is invalid.")
        sys.exit(1)
    elif operation == _IMAGE_INGESTION:
        pass
    # Retrieve paths to SDM tables, BDF files, and the NGAS staging area

    # First the pathy stuff
    if args.path:
        sdmpath = args.path
        bdfpath = args.path
        calproductpath = args.path
    else:
        # TODO This should be an error for the required flags
        sdmpath = config.getstring("archive-ingestion.SDMPath")
        bdfpath = config.getstring("archive-ingestion.BDFPath")
        calproductpath = config.getstring("archive-ingestion.BDFPath")

    if operation == _JSON_MANIFEST_INGESTION:
        operation = getoperationfromjson()

    if args.meta:  # Ingest science metadata for SDM-based observations
        filegroups = []
        if args.file:  # single file
            filegroups.append(args.file)
        else:  # all files in path
            filegroups = [x.name for x in list(Path(sdmpath).glob("*"))]
            LOG.debug(f"filesets: {filegroups}")

        groupcount = 0
        starttime = time.time()
        for filegroup in filegroups:
            filepath = Path(sdmpath) / filegroup
            if not filepath.exists():
                LOG.error(f"Observation {filepath.name} doesn't exist in {filepath.parent}")
                continue
            asdm_table = Path(filepath) / "ASDM.xml"
            if not asdm_table.exists():  # If the directory has ASDM.xml, it is a valid filegroup
                LOG.error(f"{filepath.name} is missing ASDM.xml. Not a valid observation")
                continue
            exec_block = au.getexecutionblockfromname(filegroup)
            if exec_block is not None:  # Already exists
                if args.reingest:
                    LOG.info(f"Re-ingestion requested")
                else:
                    LOG.warning(
                        f"Observation {filepath.name} already exists in metadata database and re-ingestion not requested"
                    )
                    continue
            LOG.info(f"Ingesting SDM metadata for execution block {filegroup}")
            sdm = sc.buildsdm(filegroup, sdmpath)
            if len(sdm.bdffiles) > 0:  # ingest science meta data
                project, science_products = pm.persist(sdm, args.nocal)
            else:
                LOG.error(f"Execution block: {filepath.name} has no bdfs and will not be ingested")
                continue

            LOG.info(f"SDM metadata ingestion complete for {filepath.name} in {round(time.time() - starttime, 2)}s")
            if args.stage:
                LOG.info(
                    "-t flag set. SDM tables and BDF files will be staged for NGAS ingestion in {}".format(args.stage)
                )
                ng.ingestsdmintongas(sdm, os.path.join(sdmpath, filegroup), args.stage)
            groupcount += 1

            # Send completion even to amygdala
            send_event(
                {
                    "logData": {
                        "fileset_id": os.path.basename(filegroup),
                        "ingestion_type": "observation",
                        "project_code": project.project_code,
                    },
                    "message": "ingestion complete",
                    "request": "de nada",
                }
            )

    elif args.bdf:
        # BDF Files
        # Path (-p) and file set (-o) required

        if args.nongas:
            LOG.warning("-d flag was set for no ingestion into NGAS. For bdf ingestion, this is a no-op.")
            sys.exit(0)

        if args.file and args.path:
            filelist = args.file.split(",")
            pathfilelist = []
            for file in filelist:
                filepath = os.path.join(args.path, file.strip())
                if not (filepath.endswith(".bdf") or filepath.endswith(".sdm") or filepath.endswith(".bin")):
                    LOG.warning("Invalid extension for ingestion: {}".format(filepath))
                    continue
                if not os.path.exists(filepath):
                    LOG.error("Path to file doesnt exist: {}".format(filepath))
                    continue

                pathfilelist.append(filepath)
            if len(pathfilelist) > 0:
                # Not ingesting bdfs yet
                pass
                # ng.archivefiles(pathfilelist)
            else:
                LOG.info("No files located for BDF ingestion")
        else:
            LOG.error("-b requires arguments -p and -o")

    elif args.cal:
        # Calibration products
        # Path (-p) and calibration product tar file (-o) required. Parent execution block (-g) required for calibration
        # workflow, but not bulk calibration ingestion
        if args.path and args.file and args.filegroup:
            calibration_products_file = Path(args.path) / args.file
            if not calibration_products_file.exists():
                LOG.error(
                    f"Calibration file {calibration_products_file.name} does not exist in {calibration_products_file.parent}"
                )
                sys.exit(1)
            else:
                pct.ingestcalmetadata(calibration_products_file, args.filegroup, args.nongas)
        elif args.path and args.file:
            calibration_file = Path(args.path) / args.file
            if not calibration_file.exists():
                # TODO calfile
                LOG.error(f"Path to calibration products tar file doesnt exist: {str(calibration_file)}")
                sys.exit(1)
            # Ingest file where parent execution block must be extracted from the .tar file
            pct.ingestcaltable(os.path.join(args.path, args.file))
            LOG.info(f"Calibration products {os.path.join(args.path, args.file)} ingested")
        elif args.filegroup:
            pct.ingestalmacalibration(args.filegroup)
        else:
            LOG.error("-c requires arguments -p, -o, and -g for VLA or just -g for ALMA")

    elif args.img:
        if args.path and args.filegroup:
            if not os.path.exists(args.path):
                LOG.error(f"Path to image products doesnt exist: {args.path}")
                sys.exit(1)
            LOG.info(f"Ingesting images for project {args.filegroup} in {args.path}")

            filegroup = au.getimagesetfilegroupfromname(os.path.basename(args.path))
            if filegroup is not None:  # Filegroup already exists
                if args.reingest:
                    LOG.info("Removing image set {} for re-ingestion".format(os.path.basename(args.path)))
                    au.removeimageset(filegroup)
                else:
                    LOG.warning(
                        "Image set {} already exists in the metadata database. If you wish to re-ingest, use -i -r".format(
                            os.path.basename(args.path)
                        )
                    )
                    sys.exit(1)
            else:
                LOG.info("Ingesting image set {}".format(os.path.basename(args.path)))

            pimg.ingestimagemetaData(args.path, args.filegroup, args.nongas)
        else:
            LOG.error("-i requires arguments -p and -g")
    elif args.vlba:
        if not args.path or not args.file:
            LOG.error("VLBA ingestion requires a path and file name")
            sys.exit(1)
        file_path = Path(args.path)
        fits_file = file_path / args.file
        starttime = time.time()
        existingfile = au.getvlbafile(args.file)
        if existingfile is not None:
            if args.reingest:
                # Re-ingestion - remove the existing file before ingesting
                au.removefile(existingfile)
            else:
                LOG.error(f"File: {fits_file} has already been ingested.")
                sys.exit(1)
        vlbamd.persistVLBAMetadata(fits_file)
        LOG.info(f"IDIFITS file {fits_file.name} ingested in {time.time() - starttime}s")
    elif args.json:
        if not args.path:
            LOG.critical(f"-p flag is required for ingestion via ingestion-manifest.json")
            sys.exit(1)
        ingestion_manifest_file = Path(args.path) / im.INGESTION_MANIFEST_FILE
        if not ingestion_manifest_file.is_file():
            LOG.critical(
                f"Ingestion manifest {ingestion_manifest_file.name} does not exist in {ingestion_manifest_file.parent}"
            )
            sys.exit(1)
        ingestion_manifest = IngestionManifest(ingestion_manifest_file)
        mp.ingest_from_manifest(ingestion_manifest_file.parent, ingestion_manifest)
    else:
        LOG.critical("Invalid command-line argument")
        sys.exit(1)


def getingestionoperation(args):
    if args.meta:
        return _SDM_INGESTION
    elif args.bdf:
        return _BDF_INGESTION
    elif args.cal:
        # Path (-p) and calibration product tar file (-o) required. Parent execution block (-g) required for calibration
        return _CALIBRATION_INGESTION
    elif args.img:
        return _IMAGE_INGESTION
    elif args.vlba:
        return _VLBA_IDIFITS_INGESTION
    elif args.json:
        return _JSON_MANIFEST_INGESTION
    else:
        return _INVALID_INGESTION_OPERATION


def getoperationfromjson():
    pass


def getargs(operation):
    if operation == _SDM_INGESTION:
        pass
    elif operation == _BDF_INGESTION:
        pass
    elif operation == _CALIBRATION_INGESTION:
        pass
    elif operation == _IMAGE_INGESTION:
        pass
    elif operation == _VLBA_IDIFITS_INGESTION:
        pass
    elif operation == _JSON_MANIFEST_INGESTION:
        pass


def send_event(event):
    __send_event.send(routing_key="ingestion-complete.metadata", event=event)


if __name__ == "__main__":
    main()
